# et - the trivial time tracker
# Copyright 2014 Ethan Blanton
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

require 'pathname'
require 'fcntl'
require 'json'
require 'time'
require 'date'

require 'ettask'
require 'etreport'
require 'etutil'
require 'etlog'

class ETdbBadPathError < Exception
    def initialize(full_id, tid = nil)
        @full_id = full_id
        @tid = tid
    end

    def to_s
        if @tid
            "Path #{@full_id} does not exist at #{@tid}"
        else
            "Path #{@full_id} does not exist in database"
        end
    end
end

class ETdbInvalidStateError < Exception
end

class ETdbParseError < Exception
end

class ETdb
    @@DBVER = "1.0"
    @@DUMPVER = "1.0"

    @@locked = {}

    attr_reader :UUID

    def initialize(path = Pathname.new(ENV['HOME']) + '.etdb')
        if String === path
            @dir = Pathname.new(path)
        elsif Pathname === path
            @dir = path
        else
            raise ArgumentError.new("path must be a String or Pathname")
        end

        begin
            st = File.stat(@dir)
            if not st.directory?
                raise ArgumentError.new("#{@dir} exists and is not a directory")
            end
        rescue Errno::ENOENT
            Dir.mkdir @dir
            ETdb.lock(@dir) do
                File.open(serialfile, "w") do |f|
                    f.puts("0")
                end
                File.open(tasksfile, "w") do |f|
                    f.write(JSON.pretty_generate([]))
                end
                File.open(versionfile, "w") do |f|
                    f.write(@@DBVER)
                end
                File.open(uuidfile, "w") do |f|
                    f.puts(ETUtil.uuidgen)
                end
                File.open(currentfile, "w") { }
                Dir.mkdir(logdir)
                Dir.mkdir(horizondir)
            end
        end

        @serializing = false
        @mutated = false
        @suspend_writes = false
        serialize { _reload; _load_uuid }
    end

    def add_text_task(full_id, desc)
        idchain = ETTask.idunchain full_id
        tid = idchain.pop
        serialize do
            parent = @tasks
            idchain.each do |pid|
                parent = parent.subtask pid
                if parent == nil
                    raise ETdbBadPathError.new(full_id, pid)
                end
            end
            task = ETTask.new(tid, desc)
            parent.add_subtask(task)
            _register_task task
            @mutated = true
            _write_tasks
        end
    end

    def add_task(task, parent = nil)
        serialize do
            if parent == nil
                parent = @tasks
            else
                _validate_task parent
            end
            parent.add_subtask task
            _register_task task
            if not @suspend_writes
                @mutated = true
                _write_tasks
            end
        end
    end

    def move_task(task, path)
        serialize do
            moved = false
            renamed = false
            _validate_task task
            parent = nil
            idchain = ETTask.idunchain(path)
            if idchain.length > 1
                parentpath = ETTask.idchain(idchain[(0...(idchain.length - 1))])
                parent = find_task(parentpath)
                if not parent
                    raise ETdbBadPathError(parentpath)
                end
            else
                parent = @tasks
            end

            oldparent = task.parent
            oldtid = task.tid

            # We have to remove, rename, then reinsert, because the new
            # name (if there is one) might be invalid in the old location
            # and vice-versa

            if oldparent != parent
                oldparent.remove_subtask task
            end
            if idchain.last != task.tid
                task.rename(idchain.last)
                renamed = true
            end
            if oldparent != parent
                begin
                    parent.add_subtask task
                rescue ArgumentError, ETTaskDuplicateID => a
                    task.rename(oldtid)
                    oldparent.add_subtask(task)
                    raise a
                end
                moved = true
            end
            @mutated = moved || renamed
            _write_tasks
        end
    end

    def describe_task(task, desc)
        serialize do
            _validate_task task
            task.description = desc
            @mutated = true
            _write_tasks
        end
    end

    def find_task(name)
        case name
        when /^uuid:(.*)/i
            serialize do
                task = @task_uuids[$1]
                return task
            end
        when nil
            return nil
        else
            serialize do
                idchain = ETTask.idunchain(name)
                task = @tasks
                idchain.each do |tid|
                    task = task.subtask(tid)
                    return nil if not task
                end
                return task
            end
        end
    end

    def checkin(task, time = Time.now)
        serialize do
            if not @task_uuids[task.UUID]
                raise ArgumentError.new("#{task} does not seem to be a valid task")
            end

            if @current
                raise ETdbInvalidStateError.new("Already checked in to #{@current[:task].UUID}")
            end

            @current = {
                :task => task,
                :time => time
            }
            @mutated = true
            _write_current
        end
    end

    def current
        serialize { return @current }
    end

    def cancel
        serialize do
            if @current
                @current = nil
                @mutated = true
                _write_current
            end
        end
    end

    def checkout(time = Time.now)
        serialize do
            if not @current
                raise ETdbInvalidStateError.new('Not currently checked in to any task')
            end
            task = @current[:task]
            ci = @current[:time]
            co = time
            if ci > co
                raise ETdbInvalidStateError.new("checkin (#{ci}) is after checkout (#{co})")
            end
            ETLog.new(tasklog(task)).log(ci, co)
            @current = nil
            _write_current
            @mutated = true
            return [task, ci, co]
        end
    end

    def bump(value)
        t = nil
        serialize do
            c = current
            raise ETDb.InvalidStateError if not c
            case value
            when Integer
                c[:time] += value
            when Time
                c[:time] = value
            end
            @mutated = true
            _write_current
            t = c[:time]
        end
        if t > Time.now
            $stderr.puts 'warning: bumped time is in the future!'
        end
    end

    def report(pstart, pend, options = {}, tree = nil)
        # FIXME This might be too long for our 300ms locking delay ... eventually
        if pstart > pend
            raise ArgumentError.new("Start time #{pstart} is after end time #{pend}")
        end
        serialize do
            tasks = tree ? tree : @tasks
            report = ETReport.new(tasks, options)
            tasks.each do |task|
                log = ETLog.new(tasklog(task))
                report.add_task(task, log.total(pstart, pend))
            end
            return report
        end
    end

    def dump(uuid = nil, file = $stdout)
        serialize do
            horizon = {}
            if uuid
                horizon = _load_horizon uuid
            end
            file.puts "et dump version #{@@DUMPVER}"
            file.puts "#{@tasks.subtasks.length}"
            @tasks.subtasks.each do |task|
                file.puts(JSON.generate(task.to_hash))
            end
            @tasks.each do |task|
                file.puts("task #{task.UUID}")
                begin
                    log = ETLog.new(tasklog(task))
                    log.dump(file, horizon[task.UUID])
                    horizon[task.UUID] = log.last_entry if log.last_entry
                rescue Errno::ENOENT
                end
            end
            if uuid
                _write_horizon uuid, horizon
            end
        end
    end

    def process_dump(file)
        tasks = []
        entries = Hash.new { |h,k| h[k] = [] }
        version = file.gets
        lineno = 2
        if not version =~ /^et dump version ([0-9.]+)$/
            raise ETdbParseError.new("Dump file has invalid header")
        elsif $1 != @@DUMPVER
            raise ETdbParseError.new("Expected dump version #{@@DUMPVER}, got #{$1}")
        end
        ntasks = file.gets.to_i
        lineno += 1
        ntasks.times do
            tasks << ETTask.parse(JSON.parse(file.gets))
            lineno += 1
        end
        uuid = nil
        file.each do |line|
            case line
            when /^task ([^ \n]+)$/
                uuid = $1
            else
                if not uuid
                    raise ETdbParseError.new("First task uuid not found")
                end
                values = line.rstrip.split(' ')
                if values.length != 3
                    raise ETdbParseError.new("Entry does not seem to be valid at line #{lineno}")
                end
                entries[uuid] << values
            end
            lineno += 1
        end
        serialize do
            begin
                @suspend_writes = true
                moves = []
                adds = []
                tasks.each do |task|
                    task.each do |subtask|
                        curtask = @task_uuids[subtask.UUID]
                        if not curtask
                            adds << subtask
                        else
                            if curtask.full_id != subtask.full_id
                                moves << subtask
                            end
                            if curtask.description != subtask.description
                                describe_task curtask, subtask.description
                            end
                        end
                    end
                end
                changed = 1
                while moves.length > 0 or adds.length > 0 and changed > 0
                    changed = 0
                    moved = 1
                    while moves.length > 0 and moved > 0
                        moved = 0
                        newmoves = []
                        moves.each do |task|
                            curtask = @task_uuids[task.UUID]
                            begin
                                move_task curtask, task.full_id
                            rescue
                                newmoves << task
                                next
                            end
                            moved += 1
                            changed += 1
                        end
                        moves = newmoves
                    end
                    newadds = []
                    adds.each do |task|
                        newtask = ETTask.new(task.tid, task.description, nil, [],
                                             false, task.UUID)
                        parent = nil
                        if task.parent
                            parent = @task_uuids[task.parent.UUID]
                            if not parent
                                _reload_tasks
                                raise ETdbInvalidStateError.new("New task #{task.full_id} has unknown parent")
                            end
                        end
                        begin
                            add_task(newtask, parent)
                        rescue ETTaskDuplicateID => e
                            newadds << task
                        end
                        changed += 1
                    end
                    adds = newadds
                end
                if moves.length > 0 or adds.length > 0
                    _reload_tasks
                    raise ETdbInvalidStateError.new("Task trees cannot be reconciled")
                end
                entries.each_key do |uuid|
                    if not @task_uuids[uuid]
                        raise ETdbParseError.new("Task uuid #{uuid} does not exist but has log entries")
                    end
                    log = ETLog.new(tasklog(@task_uuids[uuid]))
                    log.merge do
                        entries[uuid].each do |triplet|
                            log.try_insert(*triplet)
                        end
                    end
                end
            ensure
                @suspend_writes = false
            end
            @mutated = true
            _write_tasks
        end
    end

    def set(group, key, value)
        serialize do
            if value == nil
                if @settings[group]
                    @mutated = true if @settings[group].delete(key)
                end
            else
                if not @settings[group]
                    @settings[group] = {}
                end
                @settings[group][key] = value
                @mutated = true
            end
            _write_settings if @mutated
        end
    end

    def get(group, key)
        serialize do
            if @settings[group]
                return @settings[group][key]
            else
                return nil
            end
        end
    end

    def serial
        serialize { return @serial }
    end

    def tasks
        tasks = serialize { @tasks.subtasks }
    end

    def task(name)
        t = serialize { @tasks.subtask[name] }
    end

    def tasklog(task)
        logdir + task.UUID
    end

    def _reload
        _reload_serial
        _reload_version
        _reload_tasks
        _reload_times
        _reload_current
        _reload_settings
    end

    def _reload_serial
        File.open(serialfile, "r") do |f|
            @serial = f.gets.to_i
        end
    end

    def _reload_version
        File.open(versionfile, "r") do |f|
            ver = f.gets.strip
            if ver != @@DBVER
                raise "Unrecognized database version, aborting"
            end
        end
    end

    def _reload_tasks
        tasks = nil
        File.open(tasksfile) do |f|
            tasks = JSON.parse(f.read)
        end
        root = ETTask.new("root", "Root task")
        root.make_root
        tasks.each do |t|
            root.add_subtask(ETTask.parse(t))
        end
        @tasks = root
        # Register UUIDs
        @task_uuids = {}
        tocheck = [root]
        while not tocheck.empty?
            task = tocheck.shift
            _register_task task
            tocheck += task.subtasks
        end
    end

    def _reload_times
    end

    def _reload_current
        current = nil
        File.open(currentfile) do |f|
            data = f.read
            if data.length >= 2
                hash = JSON.parse(data) if data.length > 0
                current = {
                    :task => find_task(hash["task"]),
                    :time => Time.parse(hash["time"])
                }
                if not current[:task]
                    raise ETdbInvalidStateError.new("Current task is invalid")
                end
            end
        end
        @current = current
    end

    def _load_uuid
        File.open(uuidfile, 'r') do |f|
            @UUID = f.gets.strip
            if @UUID == ''
                raise ETdbInvalidStateError.new('UUID is empty')
            end
        end
    end

    def _load_horizon(uuid)
        begin
            File.open(horizondir + uuid, "r") do |f|
                data = f.read
                if data.length >= 2
                    return JSON.parse(data)
                end
            end
        rescue Errno::ENOENT
        end
        return {}
    end

    def _reload_settings
        begin
            File.open(settingsfile, 'r') do |f|
                data = f.read
                if data.length >= 2
                    @settings = JSON.parse(data)
                    return
                end
            end
        rescue Errno::ENOENT
        end
        @settings = {}
    end

    def _write_settings
        begin
            st = File.stat(settingsfile)
            File.rename(settingsfile, @dir + 'settings~')
        rescue Errno::ENOENT
        end
        File.open(settingsfile, 'w') do |f|
            f.write(JSON.pretty_generate(@settings))
        end
    end

    def _write_horizon(uuid, horizon)
        begin
            st = File.stat(horizondir + uuid)
            File.rename(horizondir + uuid, horizondir + (uuid + '~'))
        rescue Errno::ENOENT
        end
        File.open(horizondir + uuid, "w") do |f|
            f.write(JSON.pretty_generate(horizon))
        end
    end

    def _write_tasks
        File.rename(tasksfile, @dir + "tasks~")
        File.open(tasksfile, "w") do |f|
            f.write(JSON.pretty_generate(@tasks.subtasks.map { |t| t.to_hash }))
        end
    end

    def _write_current
        File.rename(currentfile, @dir + "current~")
        File.open(currentfile, "w") do |f|
            if @current
                data = {
                    "task" => "uuid:#{@current[:task].UUID}",
                    "time" => @current[:time].iso8601
                }
                f.write(JSON.pretty_generate(data))
            end
        end
    end

    def _register_task(task)
        if @task_uuids[task.UUID]
            raise "Task with specified UUID #{task.uuid} already exists"
        end
        @task_uuids[task.UUID] = task
    end

    def _validate_task(task)
        check = task
        while check
            break if check == @tasks
            check = check.parent
        end
        raise ETdbBadPathError.new(task.full_id) if not check
    end

    def lock
        ETdb.lock(@dir) do
            yield
        end
    end

    def serialize(&block)
        if @serializing
            return yield
        end
        ETdb.lock(@dir) do
            @serializing = true
            serial = File.open(serialfile, "r") do |f|
                f.gets.to_i
            end
            if @serial != serial
                _reload
            end
            begin
                return yield
            ensure
                if @mutated
                    @serial += 1
                    File.open(serialfile, "w") do |f|
                        f.puts(@serial)
                    end
                    @mutated = false
                end
                @serializing = false
            end
        end
    end

    def serialfile
        @dir + 'serial'
    end

    def versionfile
        @dir + 'version'
    end

    def tasksfile
        @dir + 'tasks'
    end

    def currentfile
        @dir + 'current'
    end

    def uuidfile
        @dir + 'uuid'
    end

    def settingsfile
        @dir + 'settings'
    end

    def logdir
        @dir + 'log'
    end

    def horizondir
        @dir + 'horizon'
    end

    def ETdb.lock(path)
        path = Pathname.new(path) if path.class == String

        lockfile = path + "lock"
        locked = false
        [1, 2, 3].each do
            begin
                locked = File.open(lockfile,
                                   Fcntl::O_WRONLY|Fcntl::O_CREAT|Fcntl::O_EXCL,
                                   0600) { true }
                break
            rescue Errno::EEXIST
                sleep 0.1
            end
        end
        if not locked
            raise "Could not obtain lock on #{path}"
        end
        if block_given?
            begin
                yield
            ensure
                File.unlink lockfile
            end
        else
            @@locked[lockfile.to_s] = true
        end
    end

    def ETdb.unlock(path)
        path = Pathname.new(path) if path.class == String

        lockfile = path + "lock"
        if @@locked[lockfile.to_s]
            File.unlink lockfile
            @@locked.delete lockfile.to_s
        else
            raise "Path #{path} is not locked"
        end
    end

    def ETdb.DBVER
        return @@DBVER
    end

    private :_reload, :_reload_serial, :_reload_version, :_reload_tasks,
            :_reload_times, :_reload_current, :_reload_settings, :_load_uuid,
            :_load_horizon, :_register_task, :_validate_task, :_write_tasks,
            :_write_current, :_write_horizon, :_write_settings
end
