# et - the trivial time tracker
# Copyright 2014 Ethan Blanton
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

require 'socket'

class ETLogInvalidStateError < Exception
end

class ETLog
    @@DATEFORMAT = '%Y-%m-%dT%H:%M'

    attr_reader :last_entry

    def initialize(file)
        @file = file
        @index = nil
        @f = nil
        @last_entry = nil
    end

    def dump(file, after = nil)
        started = !after ? true : false
        File.open(@file) do |f|
            f.each do |line|
                cistr, costr, gid = line.rstrip.split(' ')
                if not started
                    if gid == after
                        started = true
                    end
                    next
                end
                file.puts line
                @last_entry = gid
            end
        end
    end

    def merge(&b)
        if @index
            raise ETLogInvalidStateError.new("Recursive merge")
        end
        @index = {}
        begin
            File.open(@file, "r") do |f|
                f.each do |line|
                    cistr, costr, gid = line.rstrip.split(' ')
                    @index[gid] = true
                end
            end
        rescue Errno::ENOENT
        end
        begin
            File.open(@file, "a") do |f|
                @f = f
                yield
            end
        ensure
            @index = nil
            @f = nil
        end
    end

    def log(ci, co)
        File.open(@file, "a") do |f|
            f.puts("#{ETLog.timestamp(ci)} #{ETLog.timestamp(co)} #{ETLog.gid}")
        end
    end

    def total(from = nil, to = nil)
        t = 0
        begin
            File.open(@file, 'r') do |f|
                f.each do |line|
                    cistr, costr, gid = line.rstrip.split(' ')
                    start = Time.parse(cistr)
                    stop = Time.parse(costr)
                    if (start > stop)
                        raise ETLogInvalidStateError("Log has invalid checkin/checkout times on #{gid}")
                    end
                    if (!from || stop > from) and (!to || start < to)
                        start = from if from && start < from
                        stop = to if to && stop > to
                        t += stop.tv_sec - start.tv_sec
                    end
                end
            end
        rescue Errno::ENOENT
        end
        return t
    end

    def try_insert(cistr, costr, gid)
        if @index == nil
            raise ETLogInvalidStateError.new("Called try_insert when not inside merge")
        end
        if not @index[gid]
            @f.puts("#{cistr} #{costr} #{gid}")
        end
    end

    def self.gid
        File.open("/dev/urandom", "r") do |f|
            nonce = f.read(4).unpack('L')[0]
            "#{Socket.gethostname}-#{Time.now.iso8601}-#{@serial}-#{nonce}"
        end
    end

    def self.timestamp(time)
        time.strftime(@@DATEFORMAT)
    end
end
