# et - the trivial time tracker
# Copyright 2014 Ethan Blanton
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

require 'etutil'

class ETTaskDuplicateID < Exception
    def initialize(task, tid)
        @task = task
        @tid = tid
    end

    def to_s
        if @task != nil
            "#{@task.full_id} already has a child #{@tid}"
        else
            "{@tid} already exists"
        end
    end
end

class ETTask
    attr_reader :tid, :UUID, :parent, :subtasks, :inactive, :root
    attr_accessor :description

    @@idsep = '+'

    def initialize(tid, desc, parent = nil, subtasks = [], inactive = false, uuid = nil)
        idset = {}
        subtasks.each do |t|
            if idset[t.tid]
                raise ETTaskDuplicateID.new(nil, t.tid)
            end
            idset[t.tid] = t
            if t.parent
                raise ArgumentError("Subtask #{t.tid} already has a parent")
            end
        end
        @tid = tid
        @UUID = uuid ? uuid : ETUtil.uuidgen
        @description = desc
        @parent = parent
        @subtasks = subtasks
        @inactive = false
        @idset = idset
        @root = false
    end

    def rename(tid, desc = nil)
        if tid
            if @parent
                @parent._rename_subtask(self, @tid, tid)
            end
            @tid = tid
        end
        if desc
            @description = desc
        end
    end

    def add_subtask(task)
        if @idset[task.tid]
            raise ETTaskDuplicateID.new(self, task.tid)
        end
        if task.parent
            raise ArgumentError.new("Subtask #{task.tid} already has a parent")
        end
        @idset[task.tid] = task
        @subtasks << task
        task._setparent self
    end

    def remove_subtask(task)
        if not @idset[task.tid]
            raise ArgumentError.new("No such subtask #{task.tid}")
        end
        if task != @idset[task.tid]
            raise ArgumentError.new("Cannot delete a copy of subtask #{task.tid}")
        end
        @idset.delete task.tid
        @subtasks.delete task
        task._setparent nil
    end

    def make_root
        @root = true
    end

    def full_id
        fid = @tid
        parent = @parent
        while parent and not parent.root
            fid = "#{parent.tid}#{@@idsep}#{fid}"
            parent = parent.parent
        end
        fid
    end

    def subtask(tid)
        return @idset[tid]
    end

    def ETTask.idchain(*ids)
        return ids.join(@@idsep)
    end

    def ETTask.idunchain(ids)
        return ids.split(@@idsep)
    end

    def ETTask.parse(h)
        tid = h["tid"]
        uuid = h["uuid"]
        description = h["description"]
        subtasks = h["subtasks"]
        inactive = h["inactive"]

        if not tid or not description or not uuid or inactive == nil
            raise ArgumentError.new("Object does not seem to be an ETTask")
        end
        task = ETTask.new(tid, description, nil, [], inactive, uuid)
        if subtasks
            subtasks.each do |t|
                task.add_subtask(ETTask.parse(t))
            end
        end

        return task
    end

    def each(&b)
        tasks = [self]
        while not tasks.empty?
            task = tasks.shift
            yield task
            tasks = task.subtasks + tasks
        end
    end

    def _setparent(task)
        @parent = task
    end

    def _rename_subtask(task, oldtid, newtid)
        if @idset[newtid]
            raise ArgumentError.new("")
        end
        @idset.delete oldtid
        @idset[newtid] = task
    end

    def to_hash
        return {
            "tid" => @tid,
            "uuid" => @UUID,
            "description" => @description,
            "subtasks" => @subtasks.map { |t| t.to_hash },
            "inactive" => @inactive
        }
    end

    def depth
        d = 0
        p = parent
        while p
            d += 1
            p = p.parent
        end
        return d
    end

    def to_s
        "#{tid} #{@description}"
    end

    def describe(depth = 0)
        return subtasks.inject(sprintf("%-20s %-36s\n",
                                       "#{"  "*depth}#{tid}",
                                       description)) do |s,t|
            s + t.describe(depth + 1)
        end
    end

    protected :_setparent, :_rename_subtask
end
