require 'test/unit'

require 'fileutils'
require 'tmpdir'
require 'json'

require 'etdb'
require 'ettest'

class ETdb_Basic_Test < Test::Unit::TestCase
    include ETTest

    def setup
        @tmpdir = Pathname.new(Dir.mktmpdir)
        @dbdir = @tmpdir + 'etdb'
    end

    def teardown
        FileUtils.remove_entry @tmpdir
    end


    def test_init
        begin
            st = File.stat(@dbdir)
        rescue Exception => r
            assert(r.class == Errno::ENOENT)
        end
        assert(st == nil)

        db = ETdb.new(@dbdir)
        st = File.stat(@dbdir)
        assert st.directory?

        File.open(db.serialfile) do |f|
            assert(f.gets.to_i == 0)
        end
        st = File.stat(db.tasksfile)
        assert st.file?

        File.open(db.versionfile) do |f|
            assert(f.gets.strip == ETdb.DBVER)
        end

        st = File.stat(db.currentfile)
        assert st.file?
    end

    def test_add_task
        db = ETdb.new(@dbdir)
        db2 = ETdb.new(@dbdir)

        assert(db.serial == 0)
        assert(db.tasks.length == 0)
        assert(db.serial == 0)
        assert(db2.serial == 0)
        db.add_task(ETTask.new('TASK', 'Test task'))
        tasks = db.tasks
        assert(db.serial == 1)

        assert(tasks.length == 1)
        checktask(tasks[0], 'TASK', 'Test task', 0, nil, true)

        assert(db2.serial == 1)
        tasks = db2.tasks
        assert(tasks.length == 1)
        checktask(tasks[0], 'TASK', 'Test task', 0, nil, true)

        task = ETTask.new('TASK2', 'New Test task', nil, [], false, db.tasks[0].UUID)
        checkraise { db.add_task task }
    end

    def test_add_subtask
        db = ETdb.new(@dbdir)

        task = ETTask.new('TASK', 'Parent task')
        db.add_task(task)
        subtask = ETTask.new('SUBTASK', 'Child task')
        db.add_task(subtask, task)

        db2 = ETdb.new(@dbdir)
        assert(db2.serial == 2)
        tasks = db2.tasks

        checktask(tasks[0], 'TASK', 'Parent task', 1, nil, true)
        checktask(tasks[0].subtask('SUBTASK'), 'SUBTASK', 'Child task', 0, tasks[0], true)

        p = ETTask.new('NEWTASK', 'New task')
        c = ETTask.new('INVALID', 'Subtask with unrelated parent')
        checkraise(ETdbBadPathError) { db.add_task c, p }
        assert(db.serial == 2)
    end

    def test_add_text_tasks
        db = ETdb.new(@dbdir)

        db.add_text_task('TASK', 'Test task')
        checktask(db.tasks[0], 'TASK', 'Test task', 0, nil, true)
        db.add_text_task(ETTask.idchain('TASK', 'SUBTASK'), 'Test subtask')
        checktask(db.tasks[0].subtask('SUBTASK'), 'SUBTASK', 'Test subtask',
                  0, db.tasks[0], true)

        checkraise(ETdbBadPathError) do
            db.add_text_task(ETTask.idchain('INVALID', 'SUBTASK'),
                             'Invalid parent subtask')
        end

        assert(db.tasks.length == 1)
        assert(db.serial == 2)
    end

    def test_version
        db = ETdb.new(@dbdir)

        File.open(db.versionfile, 'w') do |f|
            f.puts 'invalid'
        end

        checkraise { db2 = ETdb.new(@dbdir) }
    end

    def test_find_task
        db = ETdb.new(@dbdir)

        t1 = ETTask.new('TASK', 'Test task')
        t2 = ETTask.new('SUBTASK', 'Test subtask')
        db.add_task(t1)
        db.add_task(t2, t1)

        assert(db.find_task('TASK') == t1)
        assert(db.find_task(ETTask.idchain('TASK', 'SUBTASK')) == t2)
        assert(db.find_task("uuid:#{t1.UUID}") == t1)
        assert(db.find_task("UUID:#{t1.UUID}") == t1)
        assert(db.find_task("uuid:#{t2.UUID}") == t2)
        assert(db.find_task('UNKNOWN') == nil)
        assert(db.find_task('uuid:unknown') == nil)
        assert(db.find_task(nil) == nil)
    end

    def test_checkin
        db = ETdb.new(@dbdir)
        t1 = ETTask.new('TASK', 'Test task')
        t2 = ETTask.new('OTHER', 'Other task')
        t3 = ETTask.new('INVALID', 'Invalid task')

        db.add_task(t1)
        db.add_task(t2)

        checkraise(ArgumentError) { db.checkin(t3) }

        assert(db.serial == 2)
        db.checkin(t1)
        assert(db.serial == 3)
        File.open(db.currentfile) do |f|
            data = JSON.parse(f.read)
            assert(data["task"] == "uuid:#{t1.UUID}")
            assert(Time.parse(data["time"]) <= Time.now)
        end

        checkraise(ETdbInvalidStateError) { db.checkin(t2) }
    end

    def test_current
        db = ETdb.new(@dbdir)
        t1 = ETTask.new('TASK', 'Test task')

        db.add_task(t1)

        assert(db.current == nil)

        db.checkin(t1)
        current = db.current
        assert(current[:task] == t1)

        db2 = ETdb.new(@dbdir)
        current = db2.current
        assert(current[:task].UUID == t1.UUID)

        File.open(db.currentfile, "w") do |f|
            f.write(JSON.pretty_generate({"time" => Time.now.iso8601, "task" => "uuid:0"}))
        end

        checkraise(ETdbInvalidStateError) { db3 = ETdb.new(@dbdir) }
    end

    def test_cancel
        db = ETdb.new(@dbdir)
        t1 = ETTask.new('TASK', 'Test task')

        db.add_task(t1)

        assert(db.serial == 1)
        db.cancel
        assert(db.serial == 1)

        db.checkin(t1)
        assert(db.serial == 2)
        db.cancel
        assert(db.serial == 3)

        st = File.stat(db.currentfile)
        assert(st.size == 0)
    end

    def test_checkout
        db = ETdb.new(@dbdir)
        t1 = ETTask.new('TASK', 'Test task')

        db.add_task(t1)

        checkraise(ETdbInvalidStateError) { db.checkout }

        db.checkin(t1)
        assert(db.serial == 2)
        db.checkout
        assert(db.serial == 3)
        st = File.stat(db.currentfile)
        assert(st.size == 0)
        assert(db.current == nil)
        st = File.stat(db.logdir + t1.UUID)
        assert(st.size > 0)

        t2 = ETTask.new('TASK2', 'Second test task')
        db.add_task(t2)
        db.checkin(t2, Time.gm(2014, 03, 06, 20, 28, 00))
        db.checkout(Time.gm(2014, 03, 06, 20, 32, 00))
        log = ETLog.new(db.tasklog(t2))
        assert(log.total == 4 * 60)
    end

    def test_move
        db = ETdb.new(@dbdir)
        t1 = ETTask.new('TASK', 'Test task')
        t2 = ETTask.new('SUBTASK', 'Test subtask')
        t3 = ETTask.new('OTHERTASK', 'Another toplevel task')

        db.add_task(t1)
        db.add_task(t2, t1)
        db.add_task(t3)

        checkraise { db.move_task(t1, 'OTHERTASK') }
        checkraise(ETTaskDuplicateID) { db.move_task(t3, ETTask.idchain('TASK', 'SUBTASK')) }
        checkraise { db.move_task(t2, ETTask.idchain('FOO', 'SUBTASK')) }

        db2 = ETdb.new(@dbdir)

        db.move_task(t2, ETTask.idchain('TASK', 'OTHERTASK'))
        assert(db2.find_task(ETTask.idchain('TASK', 'OTHERTASK')))

        db.move_task(t2, 'TOPLEVEL')
        assert(db2.find_task('TOPLEVEL'))

        db.move_task(t3, ETTask.idchain('TOPLEVEL', 'OTHERTASK'))
        assert(db2.find_task(ETTask.idchain('TOPLEVEL', 'OTHERTASK')))

        db.move_task(t1, ETTask.idchain('TOPLEVEL', 'OTHERTASK', 'THIRDTASK'))
        assert(db2.find_task(ETTask.idchain('TOPLEVEL', 'OTHERTASK', 'THIRDTASK')))

        t = db2.find_task('TOPLEVEL')
        checktask(t, 'TOPLEVEL', 'Test subtask', 1, nil, true)
        checktask(t, t2.tid, t2.description, t2.subtasks.length, nil, true)
        tt = db2.find_task(ETTask.idchain('TOPLEVEL', 'OTHERTASK'))
        checktask(tt, t3.tid, t3.description, t3.subtasks.length, t, true)
        ttt = db2.find_task(ETTask.idchain('TOPLEVEL', 'OTHERTASK', 'THIRDTASK'))
        checktask(ttt, t1.tid, t1.description, t1.subtasks.length, tt, true)

    end

    def test_describe
        db = ETdb.new(@dbdir)
        t1 = ETTask.new('TEST', 'Test task')
        t2 = ETTask.new('OTHER', 'Unrelated task')

        db.add_task t1

        checkraise(ETdbBadPathError) { db.describe_task(t2, 'New description') }

        db2 = ETdb.new(@dbdir)

        db.describe_task(t1, 'Testing task')
        assert(db2.find_task('TEST').description = 'Testing task')
    end

    def test_bump
        tmpdb = ETdb.new(@dbdir)

        time = Time.parse("2014-02-24T19:35:00-0500")

        t1 = ETTask.new("test", "testing")
        tmpdb.add_task t1

        # From here on out, tmpdb is invalid.  This next block generates
        # a task starting at a fixed time without requiring bump.
        File.open(tmpdb.currentfile, "w") do |f|
            data = {
                "task" => "uuid:#{t1.UUID}",
                "time" => time.iso8601
            }
            f.write(JSON.pretty_generate(data))
        end

        db = ETdb.new(@dbdir)
        assert(db.current[:time].tv_sec == time.tv_sec)

        db.bump(-60)
        newtime = Time.parse("2014-02-24T19:34:00-0500")
        assert(db.current[:time].tv_sec == newtime.tv_sec)

        db.bump(60)
        assert(db.current[:time].tv_sec == time.tv_sec)

        newtime = Time.parse("1970-01-01T00:00:01Z")
        db.bump(newtime)
        assert(db.current[:time].tv_sec == newtime.tv_sec)
    end

    def test_uuid
        db = ETdb.new(@dbdir)

        File.open(db.uuidfile) do |f|
            uuid = f.gets.strip
            assert(uuid =~ /^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$/)
        end
    end
end

